/** options for each client instance */
interface ClientOptions extends RequestInit {
    /** set the common root URL for all API requests */
    baseUrl?: string;
}
export interface BaseParams {
    params?: {
        query?: Record<string, unknown>;
    };
}
export type HttpMethod = 'get' | 'put' | 'post' | 'delete' | 'options' | 'head' | 'patch' | 'trace';
export type OkStatus = 200 | 201 | 202 | 203 | 204 | 206 | 207;
export type ErrorStatus = 500 | 400 | 401 | 402 | 403 | 404 | 405 | 406 | 407 | 408 | 409 | 410 | 411 | 412 | 413 | 414 | 415 | 416 | 417 | 418 | 420 | 421 | 422 | 423 | 424 | 425 | 426 | 429 | 431 | 444 | 450 | 451 | 497 | 498 | 499;
/** Get a union of paths which have method */
export type PathsWith<Paths extends {}, PathnameMethod extends HttpMethod> = {
    [Pathname in keyof Paths]: Paths[Pathname] extends {
        [K in PathnameMethod]: any;
    } ? Pathname : never;
}[keyof Paths];
/** Find first match of multiple keys */
export type FilterKeys<Obj, Matchers> = {
    [K in keyof Obj]: K extends Matchers ? Obj[K] : never;
}[keyof Obj];
/** handle "application/json", "application/vnd.api+json", "appliacation/json;charset=utf-8" and more */
export type JSONLike = `${string}json${string}`;
export type Params<OperationObj> = OperationObj extends {
    parameters: any;
} ? {
    params: OperationObj['parameters'];
} : BaseParams;
export type RequestBodyObj<OperationObj> = FilterKeys<OperationObj, 'requestBody'>;
export type RequestBodyContent<OperationObj> = undefined extends RequestBodyObj<OperationObj> ? FilterKeys<NonNullable<RequestBodyObj<OperationObj>>, 'content'> | undefined : FilterKeys<RequestBodyObj<OperationObj>, 'content'>;
export type RequestBodyJSON<OperationObj> = FilterKeys<RequestBodyContent<OperationObj>, JSONLike> extends never ? FilterKeys<NonNullable<RequestBodyContent<OperationObj>>, JSONLike> | undefined : FilterKeys<RequestBodyContent<OperationObj>, JSONLike>;
export type RequestBody<OperationObj> = undefined extends RequestBodyJSON<OperationObj> ? {
    body?: RequestBodyJSON<OperationObj>;
} : {
    body: RequestBodyJSON<OperationObj>;
};
export type QuerySerializer<OperationObj> = {
    querySerializer?: (query: OperationObj extends {
        parameters: {
            query: any;
        };
    } ? OperationObj['parameters']['query'] : Record<string, unknown>) => string;
};
export type FetchOptions<OperationObj> = Params<OperationObj> & RequestBody<OperationObj> & Omit<RequestInit, 'body'> & QuerySerializer<OperationObj>;
export type Success<OperationObj> = FilterKeys<FilterKeys<OperationObj, OkStatus>, 'content'>;
export type Error<OperationObj> = FilterKeys<FilterKeys<OperationObj, ErrorStatus>, 'content'>;
export type FetchResponse<T> = {
    data: T extends {
        responses: any;
    } ? NonNullable<FilterKeys<Success<T['responses']>, JSONLike>> : unknown;
    error?: never;
    response: Response;
} | {
    data?: never;
    error: T extends {
        responses: any;
    } ? NonNullable<FilterKeys<Error<T['responses']>, JSONLike>> : unknown;
    response: Response;
};
export default function createClient<Paths extends {}>(options?: ClientOptions): {
    /** Call a GET endpoint */
    get<Pathname extends PathsWith<Paths, "get">>(url: Pathname, init: FetchOptions<FilterKeys<Paths[Pathname], "get">>): Promise<FetchResponse<Paths[Pathname][keyof Paths[Pathname]]>>;
    /** Call a PUT endpoint */
    put<Pathname_1 extends PathsWith<Paths, "put">>(url: Pathname_1, init: FetchOptions<FilterKeys<Paths[Pathname_1], "put">>): Promise<FetchResponse<Paths[Pathname_1][keyof Paths[Pathname_1]]>>;
    /** Call a POST endpoint */
    post<Pathname_2 extends PathsWith<Paths, "post">>(url: Pathname_2, init: FetchOptions<FilterKeys<Paths[Pathname_2], "post">>): Promise<FetchResponse<Paths[Pathname_2][keyof Paths[Pathname_2]]>>;
    /** Call a DELETE endpoint */
    del<Pathname_3 extends PathsWith<Paths, "delete">>(url: Pathname_3, init: FetchOptions<FilterKeys<Paths[Pathname_3], "delete">>): Promise<FetchResponse<Paths[Pathname_3][keyof Paths[Pathname_3]]>>;
    /** Call a OPTIONS endpoint */
    options<Pathname_4 extends PathsWith<Paths, "options">>(url: Pathname_4, init: FetchOptions<FilterKeys<Paths[Pathname_4], "options">>): Promise<FetchResponse<Paths[Pathname_4][keyof Paths[Pathname_4]]>>;
    /** Call a HEAD endpoint */
    head<Pathname_5 extends PathsWith<Paths, "head">>(url: Pathname_5, init: FetchOptions<FilterKeys<Paths[Pathname_5], "head">>): Promise<FetchResponse<Paths[Pathname_5][keyof Paths[Pathname_5]]>>;
    /** Call a PATCH endpoint */
    patch<Pathname_6 extends PathsWith<Paths, "patch">>(url: Pathname_6, init: FetchOptions<FilterKeys<Paths[Pathname_6], "patch">>): Promise<FetchResponse<Paths[Pathname_6][keyof Paths[Pathname_6]]>>;
    /** Call a TRACE endpoint */
    trace<Pathname_7 extends PathsWith<Paths, "trace">>(url: Pathname_7, init: FetchOptions<FilterKeys<Paths[Pathname_7], "trace">>): Promise<FetchResponse<Paths[Pathname_7][keyof Paths[Pathname_7]]>>;
};
export {};
