// settings
const DEFAULT_HEADERS = {
    'Content-Type': 'application/json',
};
export default function createClient(options) {
    const defaultHeaders = new Headers({
        ...DEFAULT_HEADERS,
        ...(options?.headers ?? {}),
    });
    async function coreFetch(url, fetchOptions) {
        let { headers, body: requestBody, params = {}, querySerializer = (q) => new URLSearchParams(q).toString(), ...init } = fetchOptions || {};
        // URL
        let finalURL = `${options?.baseUrl ?? ''}${url}`;
        if (params.path) {
            for (const [k, v] of Object.entries(params.path))
                finalURL = finalURL.replace(`{${k}}`, encodeURIComponent(String(v)));
        }
        if (params.query && Object.keys(params.query).length) {
            finalURL += `?${querySerializer(params.query)}`;
        }
        // headers
        const baseHeaders = new Headers(defaultHeaders); // clone defaults (don’t overwrite!)
        const headerOverrides = new Headers(headers);
        for (const [k, v] of headerOverrides.entries()) {
            if (v === undefined || v === null)
                baseHeaders.delete(k); // allow `undefined` | `null` to erase value
            else
                baseHeaders.set(k, v);
        }
        // fetch!
        const response = await fetch(finalURL, {
            redirect: 'follow',
            ...options,
            ...init,
            headers: baseHeaders,
            body: typeof requestBody === 'string' ? requestBody : JSON.stringify(requestBody),
        });
        // don’t parse JSON if status is 204, or Content-Length is '0'
        const body = response.status === 204 || response.headers.get('Content-Length') === '0' ? {} : await response.json();
        return response.ok ? { data: body, response } : { error: body, response: response };
    }
    return {
        /** Call a GET endpoint */
        async get(url, init) {
            return coreFetch(url, { ...init, method: 'GET' });
        },
        /** Call a PUT endpoint */
        async put(url, init) {
            return coreFetch(url, { ...init, method: 'PUT' });
        },
        /** Call a POST endpoint */
        async post(url, init) {
            return coreFetch(url, { ...init, method: 'POST' });
        },
        /** Call a DELETE endpoint */
        async del(url, init) {
            return coreFetch(url, { ...init, method: 'DELETE' });
        },
        /** Call a OPTIONS endpoint */
        async options(url, init) {
            return coreFetch(url, { ...init, method: 'OPTIONS' });
        },
        /** Call a HEAD endpoint */
        async head(url, init) {
            return coreFetch(url, { ...init, method: 'HEAD' });
        },
        /** Call a PATCH endpoint */
        async patch(url, init) {
            return coreFetch(url, { ...init, method: 'PATCH' });
        },
        /** Call a TRACE endpoint */
        async trace(url, init) {
            return coreFetch(url, { ...init, method: 'TRACE' });
        },
    };
}
